import asyncio
import glob
import hashlib
import logging
import os
import re
import shlex
import typing
from . import transformations


LOGGER = logging.getLogger(__name__)


class Artifact:
    """
    Artifact generated by a transformation.

    Args:
        name: Unique name of the artifact.

    Raises:
        ValueError: If the given :attr:`name` is already in use by another artifact.

    Attributes:
        children: Transformations that consume this artifact; empty if the artifact is a leaf of the
            directed acyclic graph.
        REGISTRY: Mapping of artifact names to :class`Artifact` instances. The registry is used to
            keep track of all artifacts and ensure artifact names are unique.
    """
    def __init__(self, name: str) -> None:
        if self.REGISTRY.setdefault(name, self) is not self:
            raise ValueError(f"artifact with name `{name}` already exists")
        self.name = name
        self._parent = None
        self.children = []

    children: typing.Iterable[transformations.Transformation]

    @property
    def parent(self) -> 'transformations.Transformation':
        """
        Transformation that generates this artifact; :code:`None` if the artifact is a root of the
        directed acyclic graph.
        """
        return self._parent

    @parent.setter
    def parent(self, value: 'transformations.Transformation') -> None:
        if self._parent is not None:
            raise RuntimeError(f"{self} is already associated with a transformation")
        self._parent = value

    @property
    def digest(self) -> bytes:
        """
        Concise summary of the artifact; :code:`None` if the artifact does not exist, cannot be
        summarized, or should always be generated using its :attr:`parent` transform.
        """
        return None

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    async def __call__(self):
        if self.parent is None:
            return
        return await self.parent

    def __await__(self):
        # See https://stackoverflow.com/a/57078217/1150961 for details.
        return (yield from self().__await__())

    REGISTRY: typing.Mapping[str, "Artifact"] = {}


class File(Artifact):
    """
    Artifact representing a file.
    """
    def __init__(self, name: str) -> None:
        super().__init__(name)
        if re.search(r"\s", self.name):
            LOGGER.warning("whitespace in `%s` is a recipe for disaster; expect the unexpected",
                           self.name)
        self._last_modified = None
        self._digest = None

    async def __call__(self):
        await super().__call__()
        if not os.path.exists(self.name):
            if self._parent:
                message = f"{self._parent} did not generate {self}"
            else:
                message = f"{self} does not exist and cannot be generated"
            raise FileNotFoundError(message)

    @property
    def digest(self):
        try:
            # Only re-evaluate the digest if the file has not been modified since we last computed
            # the hash.
            last_modified = os.stat(self.name).st_mtime
            if self._last_modified is None or self._last_modified < last_modified:
                algorithm = hashlib.sha256()
                with open(self.name, 'rb') as fp:
                    while (chunk := fp.read(4096)):
                        algorithm.update(chunk)
                self._digest = algorithm.digest()
                self._last_modified = last_modified
            return self._digest
        except FileNotFoundError:
            return None

    @classmethod
    def glob(self, pattern: str) -> typing.Iterable["File"]:
        """
        Create a list of file artifacts based on a glob pattern.

        Args:
            pattern: Pattern passed to :func:`glob.glob`.

        Returns:
            artifacts: File artifacts.
        """
        return normalize_artifacts([filename for filename in glob.glob(pattern)])

    def __str__(self):
        return shlex.quote(self.name)


def normalize_artifacts(
    artifacts: typing.Union[str, Artifact, typing.Iterable[typing.Union[str, Artifact]]]) \
        -> typing.Iterable[Artifact]:
    r"""
    Normalize one or more artifacts. Strings are implicitly converted to :class:`FileArtifact`\ s.

    Args:
        artifacts: One or more artifacts.

    Returns:
        normalized: List of normalized artifacts.

    Raises:
        ValueError: If an artifact that is not a :class:`File` is referred to by name only.
    """
    if artifacts is None:
        return []
    if not isinstance(artifacts, typing.Iterable) or isinstance(artifacts, str):
        artifacts = [artifacts]
    normalized = []
    for artifact in artifacts:
        if isinstance(artifact, str):
            try:
                candidate = Artifact.REGISTRY[artifact]
                if isinstance(candidate, File):
                    artifact = candidate
                else:
                    raise ValueError("only file artifacts can be implicitly referenced by name")
            except KeyError:
                artifact = File(artifact)
        elif not isinstance(artifact, Artifact):
            raise TypeError(artifact)
        normalized.append(artifact)
    return normalized


async def gather_artifacts(*artifacts_or_transformations, num_concurrent: int = None) \
       -> typing.Coroutine:
    """
    Gather one or more artifacts and wrap them in a coroutine.

    Args:
        *artifacts_or_transformations: Artifacts to gather and/or transformations whose artifacts to
            gather.
        num_concurrent: Maximum number of concurrent transformations. Defaults to unrestricted.

    Returns:
        gathered: Coroutine that awaits all gathered artifacts.
    """
    try:
        if num_concurrent is not None:
            transformations.Transformation.SEMAPHORE = asyncio.Semaphore(num_concurrent)
        gathered = []
        for item in artifacts_or_transformations:
            if isinstance(item, Artifact):
                gathered.append(item)
            elif isinstance(item, transformations.Transformation):
                gathered.extend(item)
            else:
                raise TypeError(item)
        await asyncio.gather(*gathered)
    finally:
        transformations.Transformation.SEMAPHORE = None
